# -*- coding: UTF-8 -*-
"""
@Summary : docstr
@Author  : Rey
@Time    : 2022-04-02 10:24:24
"""
import math

from django.http import HttpRequest, JsonResponse
from django.utils import timezone
from django.views.decorators.http import require_GET

from notice.settings import NOTICE_ALLOWED_TYPED_CLASS
from notice.models import NoticeStore, ReceiverTag
from notice.response import AuthFailed, NotFound, ValidationFailed, ValidationFailedDetailEnum


def get_page_notice(receiver_id, page, size, title=None, judge_kwargs: dict =dict()):
    allowed_notice_type_ids, allowed_receiver_type_ids = NOTICE_ALLOWED_TYPED_CLASS(receiver_id=receiver_id, judge_kwargs=judge_kwargs).judge()
    if not allowed_notice_type_ids or not allowed_receiver_type_ids:
        return JsonResponse(data={
            'total': 0,
            'max_page': 1,
            'page': page,
            'items': []
        })

    filter_params = {
        'is_draft': False,
        'publish_at__lte': timezone.now(),
        'receiver_type_ids__overlap': allowed_receiver_type_ids,
        'notice_type_id__in': allowed_notice_type_ids,
    }
    if title:
        filter_params['title__contains'] = title
    total = NoticeStore.objects.filter(**filter_params).count()
    max_page = math.ceil(total / size)

    items = [
        {
            'id': item.id,
            'title': item.title,
            'publish_at': item.published_at,
            'is_read': True if hasattr(item, 'receivertag') else False,
        }
        for item in NoticeStore.objects.filter(
            **filter_params
        ).only(
            'title', 'publish_at', 'receivertag__id', 'is_draft'
        ).order_by('-id')[(page-1)*size: page*size]
    ] if page <= max_page else []

    return JsonResponse(data={
        'total': total,
        'max_page': max_page,
        'page': page,
        'items': items
    })


@require_GET
def list_notice(request: HttpRequest):
    if not request.user.is_authenticated:
        return AuthFailed()

    params = request.GET
    page = params.get('page', '1')
    if not page.isdigit():
        return ValidationFailed(ValidationFailedDetailEnum.PAGE.value)
    page = int(page)

    size = params.get('size', '10')
    if not size.isdigit():
        return ValidationFailed(ValidationFailedDetailEnum.SIZE.value)
    size = int(size)

    title = params.get('title', '')

    return get_page_notice(request.user.pk, page, size, title)


def retrieve_notice(receiver_id, pk, judge_kwargs: dict = dict()):
    allowed_notice_type_ids, allowed_receiver_type_ids = NOTICE_ALLOWED_TYPED_CLASS(receiver_id, judge_kwargs).judge()
    if not allowed_notice_type_ids or not allowed_receiver_type_ids:
        return NotFound()

    filter_params = {
        'is_draft': False,
        'publish_at__lte': timezone.now(),
        'receiver_type_ids__overlap': allowed_receiver_type_ids,
        'notice_type_id__in': allowed_notice_type_ids,
        'pk': pk
    }
    notice = NoticeStore.objects.filter(**filter_params).only(
        'publish_at', 'title', 'content', 'is_draft'
    ).first()
    if not notice:
        return NotFound()

    if not ReceiverTag.objects.filter(
        receiver_id=receiver_id, noticestore_id=pk
    ).exists():
        ReceiverTag.objects.create(
            receiver_id=receiver_id,
            noticestore_id=pk,
            read_at=timezone.now()
        )

    resp = {
        'id': notice.id,
        'title': notice.title,
        'content': notice.content,
        'publish_at': notice.published_at,
    }
    return JsonResponse(data=resp)


@require_GET
def some_notice(request: HttpRequest, pk: int):
    if not request.user.is_authenticated:
        return AuthFailed()

    return retrieve_notice(request.user.pk, pk)



# Generated by Django 4.0.3 on 2022-04-08 06:14

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notice', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='noticestore',
            name='receiver_type',
        ),
        migrations.AddField(
            model_name='noticestore',
            name='receiver_type_ids',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.IntegerField(), null=True, size=None, verbose_name='receiver type'),
        ),
    ]



# -*- coding: UTF-8 -*-
"""
@Summary : views
@Author  : Rey
@Time    : 2022-04-02 18:18:19
"""
from rest_framework.views import APIView

from notice.views.admin import notice as notice_view


class ListNoticeView(APIView):
    def get(self, request, *args, **kwargs):
        return notice_view(request)



# -*- coding: UTF-8 -*-
"""
@Summary : admin views
@Author  : Rey
@Time    : 2022-03-30 13:16:11
"""
import json
import math

from django.http import JsonResponse, HttpRequest
from django.utils import timezone
from django.views.decorators.http import require_http_methods

from notice.forms import NoticeForm, ChangeTimingForm
from notice.models import NoticeStore, NoticeType, ReceiverType
from notice.response import AuthFailed, NotFound, ValidationFailed, ValidationFailedDetailEnum
from notice.settings import NOTICE_DATETIME_FORMAT


def list_all_notice_types(request):
    notice_types = NoticeType.objects.all().only('id', 'desc').order_by('id')
    data = [{'id': item.id, 'desc': item.desc} for item in notice_types]
    return JsonResponse(data=data, safe=False)


def list_all_receiver_types(request):
    receiver_types = ReceiverType.objects.all().only('id', 'desc').order_by('id')
    data = [{'id': item.id, 'desc': item.desc} for item in receiver_types]
    return JsonResponse(data=data, safe=False)


def create_notice(data: dict, creator_id: int):
    # TODO(Rey): if send way is now or timing, all fields are required.
    f = NoticeForm(data)
    if not f.is_valid():
        return ValidationFailed(f.errors)

    data = f.cleaned_data
    data.pop('send_way')
    data['notice_type_id'] = data.pop('type_id')
    data['creator_id'] = creator_id

    notice = NoticeStore.objects.create(**data)
    return JsonResponse(data={'id': notice.pk})


def list_notice(params: dict):
    page = params.get('page', '1')
    if not page.isdigit():
        return ValidationFailed(ValidationFailedDetailEnum.PAGE.value)
    page = int(page)

    size = params.get('size', '10')
    if not size.isdigit():
        return ValidationFailed(ValidationFailedDetailEnum.SIZE.value)
    size = int(size)

    total = NoticeStore.objects.all().count()
    max_page = math.ceil(total / size)

    items = [
        {
            'id': item.id,
            'title': item.title,
            'created_at': item.created_at.strftime(NOTICE_DATETIME_FORMAT),
            'publish_at': item.publish_at.strftime(NOTICE_DATETIME_FORMAT) if item.publish_at else None,
            'type': item.notice_type.desc,
            'status': item.status.value,
            'status_desc': NoticeStore.StatusLabel.get(item.status),
        }
        for item in NoticeStore.objects.all().select_related(
            'notice_type'
        ).order_by('-id')[(page-1)*size: page*size]
    ] if page <= max_page else []

    resp = {
        'total': total,
        'max_page': max_page,
        'page': page,
        'items': items
    }
    return JsonResponse(data=resp)


@require_http_methods(['GET', 'POST'])
def notice(request: HttpRequest):
    if request.method == 'POST':
        if not request.user.is_authenticated:
            return AuthFailed()
        return create_notice(json.loads(request.body), request.user.pk)
    params = request.GET
    return list_notice(params)


def retrieve_notice(pk: int):
    notice = NoticeStore.objects.filter(pk=pk).only(
        'is_draft', 'publish_at', 'title', 'content', 'notice_type_id', 'receiver_type_ids',
    ).first()
    if not notice:
        return NotFound()

    resp = {
        'id': notice.id,
        'title': notice.title,
        'content': notice.content,
        'publish_at': notice.published_at,
        'type_id': notice.notice_type_id,
        'receiver_type_ids': notice.receiver_type_ids,
    }
    return JsonResponse(data=resp)


def delete_notice(pk: int):
    if not NoticeStore.objects.filter(pk=pk).exists():
        return NotFound()

    notice = NoticeStore.objects.get(pk=pk)
    if notice.status != NoticeStore.StatusEnum.DRAFT:
        return ValidationFailed(ValidationFailedDetailEnum.DELETE_NOT_DRAFT.value)

    if notice.status == NoticeStore.StatusEnum.DONE:
        return ValidationFailed(ValidationFailedDetailEnum.DELETE_PUBLISHED.value)

    NoticeStore.objects.filter(pk=pk).update(
        is_deleted=True,
        updated_at=timezone.now()
    )

    return JsonResponse(data={})


def put_notice(data: dict, pk: int):
    notice = NoticeStore.objects.filter(pk=pk).first()
    if not notice:
        return NotFound()

    if notice.status != NoticeStore.StatusEnum.DRAFT:
        return ValidationFailed(ValidationFailedDetailEnum.CHANGE_NOT_DRAFT.value)

    f = NoticeForm(data)
    if not f.is_valid():
        return ValidationFailed(f.errors)

    data = f.cleaned_data
    data['notice_type_id'] = data.pop('type_id')

    update_params = {}
    for key, value in data.items():
        if hasattr(notice, key) and value != getattr(notice, key):
            update_params[key] = value
    if update_params:
        update_params['updated_at'] = timezone.now()
        NoticeStore.objects.filter(pk=pk).update(**update_params)
    return JsonResponse(data={})


@require_http_methods(['GET', 'PUT', 'DELETE'])
def some_notice(request: HttpRequest, pk: int):
    if request.method == 'DELETE':
        return delete_notice(pk)
    if request.method == 'PUT':
        return put_notice(json.loads(request.body), pk)
    return retrieve_notice(pk)


def change_timing(data: dict, pk: int):
    if not NoticeStore.objects.filter(pk=pk).exists():
        return NotFound()

    notice: NoticeStore = NoticeStore.objects.get(pk=pk)
    if notice.status == NoticeStore.StatusEnum.DRAFT:
        return ValidationFailed(ValidationFailedDetailEnum.CHANGE_DRAFT_TIMING.value)
    if notice.status == NoticeStore.StatusEnum.DONE:
        return ValidationFailed(ValidationFailedDetailEnum.CHANGE_PUBLISHED.value)

    f = ChangeTimingForm(data=data)
    if not f.is_valid():
        return ValidationFailed(f.errors)

    NoticeStore.objects.filter(pk=pk).update(
        publish_at=f.cleaned_data['publish_at'],
        updated_at=timezone.now()
    )
    return JsonResponse(data={})


def delete_timing(pk):
    """取消定时发送"""
    if not NoticeStore.objects.filter(pk=pk).exists():
        return NotFound()

    notice: NoticeStore = NoticeStore.objects.get(pk=pk)
    if notice.status == NoticeStore.StatusEnum.DRAFT:
        return ValidationFailed(ValidationFailedDetailEnum.DELETE_DRAFT_TIMING.value)
    if notice.status == NoticeStore.StatusEnum.DONE:
        return ValidationFailed(ValidationFailedDetailEnum.CHANGE_PUBLISHED.value)

    NoticeStore.objects.filter(pk=pk).update(
        is_draft=True,
        updated_at=timezone.now(),
        publish_at=None
    )
    return JsonResponse(data={})


@require_http_methods(['PUT', 'DELETE'])
def change_timing_notice(request: HttpRequest, pk: int):
    if request.method == 'PUT':
        data = json.loads(request.body)
        return change_timing(data, pk)
    return delete_timing(pk)



# -*- coding: UTF-8 -*-
"""
@Summary : docstr
@Author  : Rey
@Time    : 2022-04-02 10:58:14
"""
from notice.helpers import BaseGetAllowedTypes


class ExampleGetAllowedViewTypes(BaseGetAllowedTypes):

    def judge_notice_types(self) -> list:
        all_notice_types = self.get_all_notice_type_names()
        allowed_notice_type_ids = []

        self.done_notice_type.append('system')
        if self.receiver_id:
            allowed_notice_type_ids.append(all_notice_types['system'])

        self.done_notice_type.append('private')
        if self.receiver_id:
            allowed_notice_type_ids.append(all_notice_types['private'])

        if set(all_notice_types) - set(self.done_notice_type) != set():
            raise NotImplementedError(
                'not judge all existed notice types: to judge={}'.format(set(all_notice_types) - set(self.done_notice_type))
            )
        return allowed_notice_type_ids

    def judge_notice_receiver_types(self) -> list:
        all_receiver_types = self.get_all_receiver_type_names()
        allowed_receiver_type_ids = []

        self.done_receiver_type.append('all')
        if self.receiver_id:
            allowed_receiver_type_ids.append(all_receiver_types['all'])

        self.done_receiver_type.append('part')
        # if self.receiver_id:
        #     allowed_receiver_type_ids.append(all_receiver_types['part'])

        if set(all_receiver_types) - set(self.done_receiver_type) != set():
            raise NotImplementedError(
                'not judge all existed receiver types: to judge={}'.format(set(all_receiver_types) - set(self.done_receiver_type))
            )
        return allowed_receiver_type_ids



# -*- coding: UTF-8 -*-
"""
@Summary : setup
@Author  : Rey
@Time    : 2022-04-02 14:00:57
"""
from notice import get_version
from setuptools import find_packages, setup


install_requires = ['Django >= 2.2',]

setup(
    name='django-simple-notice',
    version=get_version(),
    description='A very simple, yet powerful, Django notice application',
    long_description='A very simple, yet powerful, Django notice application',
    author='rey',
    author_email="zpcao.first@google.com",
    url='https://github.com/czp-first/django-simple-notice',
    license="MIT",
    packages=find_packages(exclude=["testproject", "testproject.*"]),
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Security",
        "Topic :: Internet :: WWW/HTTP",
        "Framework :: Django",
    ],
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
)


# -*- coding: UTF-8 -*-
"""
@Summary : response and error msg
@Author  : Rey
@Time    : 2022-04-03 11:09:46
"""
from enum import Enum

from django.core.serializers.json import DjangoJSONEncoder
from django.http import JsonResponse
from django.utils.translation import gettext_lazy as _


class BaseFailedJsonResponse(JsonResponse):
    status_code = 400
    default_detail = 'bad request'
    default_code = 'error'

    def __init__(self,
        detail=None,
        code=None,
        encoder=DjangoJSONEncoder,
        safe=True,
        json_dumps_params=None,
        **kwargs,):
        if detail is None:
            detail = self.default_detail
        if code is None:
            code = self.default_code
        super().__init__({'detail': detail, 'code': code}, encoder, safe, json_dumps_params, **kwargs)



class ValidationFailed(BaseFailedJsonResponse):
    status_code = 400
    default_detail = 'invalid input'
    default_code = 'invalid'


class AuthFailed(BaseFailedJsonResponse):
    status_code = 401
    default_detail = 'Auth Failed'
    default_code = 'authentication_failed'


class NotFound(BaseFailedJsonResponse):
    status_code = 404
    default_detail = 'Not Found'
    default_code = 'not_found'


class ValidationFailedDetailEnum(Enum):
    NOTICE_TYPE = _('Invalid Notice Type')
    RECEIVER_TYPE = _('Invalid Receiver Type')
    SEND_WAY = _('Invalid Send Way')
    PUBLISH_TIME = _('Invalid Publish Time')
    PAGE = _('Invalid Page')
    SIZE = _('Invalid Size')

    OUTDATE = _('Cant Set Time Which Is Out Of Date')
    CHANGE_NOT_DRAFT = _('Cant Change Notice Which Is Not Draft')
    CHANGE_DRAFT_TIMING = _('Cant Change Timing Which Is Draft')
    DELETE_DRAFT_TIMING = _('Cant Delete Timing Which Is Draft')
    DELETE_NOT_DRAFT = _('Cant Delete Notice Which Is Not Draft')
    CHANGE_PUBLISHED = _('Cant Change Notice Which Has Been published')
    DELETE_PUBLISHED = _('Cant Delete Notice Which Has Been published')



# -*- coding: UTF-8 -*-
"""
@Summary : docstr
@Author  : Rey
@Time    : 2022-04-02 10:58:14
"""
from abc import ABCMeta, abstractmethod

from django.apps import apps


class BaseGetAllowedTypes(metaclass=ABCMeta):
    """include receiver types and notice types"""
    def __init__(self, receiver_id: int, judge_kwargs: dict = dict()) -> None:
        if not isinstance(receiver_id, int):
            raise TypeError('invalid type: receiver user id')
        self.receiver_id = receiver_id
        self.done_receiver_type = list()
        self.done_notice_type = list()
        if judge_kwargs:
            for key, value in judge_kwargs.items():
                setattr(self, key, value)

    @staticmethod
    def get_all_notice_type_names():
        return dict(apps.get_model('notice.NoticeType').objects.all().values_list('name', 'id'))

    @staticmethod
    def get_all_receiver_type_names():
        return dict(apps.get_model('notice.ReceiverType').objects.all().values_list('name', 'id'))

    @abstractmethod
    def judge_notice_receiver_types(self) -> list:
        pass

    @abstractmethod
    def judge_notice_types(self) -> list:
        pass

    def judge(self):
        return self.judge_notice_types(), self.judge_notice_receiver_types()



# -*- coding: UTF-8 -*-
"""
@Summary : app的配置
@Author  : Rey
@Time    : 2022-03-30 11:58:02
"""
from importlib import import_module

from django.conf import settings

from notice.helpers import BaseGetAllowedTypes


NOTICE_CREATOR_MODEL = getattr(settings, 'NOTICE_CREATOR_MODEL', settings.AUTH_USER_MODEL)
NOTICE_RECEIVER_MODEL = getattr(settings, 'NOTICE_RECEIVER_MODEL', settings.AUTH_USER_MODEL)
NOTICE_DATETIME_FORMAT = getattr(settings, 'NOTICE_DATETIME_FORMAT', '%Y-%m-%d %H:%M:%S')

def get_notice_allowed_types_cls():
    cls_conf = getattr(settings, 'NOTICE_ALLOWED_TYPED_CLASS')
    if not cls_conf or not isinstance(cls_conf, dict) or 'module' not in cls_conf or 'class' not in cls_conf:
        raise NotImplementedError('please conf notice allowed types class correctly!')
    module = import_module(cls_conf['module'])
    cls = getattr(module, cls_conf['class'])
    if not issubclass(cls, BaseGetAllowedTypes):
        raise NotImplementedError('please implement notice allowed types class correctly!')
    return cls

NOTICE_ALLOWED_TYPED_CLASS = get_notice_allowed_types_cls()



# -*- coding: UTF-8 -*-
"""
@Summary : urls
@Author  : Rey
@Time    : 2022-04-02 17:12:23
"""
from django.urls import path

from notice.views import admin as admin_views
from notice.views import client as client_views

admin_urlpatterns = [
    path('admin/', admin_views.notice, name='admin-notice'),
    path('admin/<int:pk>/', admin_views.some_notice, name='admin-some-notice'),
    path('admin/receiver_types/', admin_views.list_all_receiver_types, name='admin-receiver-types'),
    path('admin/types/', admin_views.list_all_notice_types, name='admin-notice-types'),
    path('admin/timing/<int:pk>/', admin_views.change_timing_notice, name='admin-change-timing-notice'),
]

client_urlpatterns = [
    path('client/', client_views.list_notice, name='client-list-notice'),
    path('client/<int:pk>/', client_views.some_notice, name='client-retrieve-notice'),
]

urlpatterns = admin_urlpatterns + client_urlpatterns
